module Automaton

"""
Creates and initializes state automaton having the state transition function
`stf` and initial state `x0`.

State transition function should be a mapping of the form `(x, u) -> (x_, y)`
where `x` and `u` are the current state and input, `x_` is the consequent state
and `y` is the resulting output.

For performance reasons, `stf` is *not* assumed to be a pure function! The state
may be mutated inline. Thus it is not safe to directly alter `x_`. However, it
should be safe to directly alter `y`. All consequent implementations must take
theses facts into consideration.

Returns the corresponding *automaton*, that is a function `u -> (y, [x_])` with
embedded state. Each consequent invocation of this function evaluates output and
advances the state. Deep copy of the state is returned, in addition to output,
only if `exposeState` parameter is set to `true`.
"""
function toInOut(stf, x0; exposeState=false)
    # Use array so that `x` is actually a `pointer` to state, not the state
    # itself. This is important because the content of `x` will be modified
    # by consequent invocations of the inner function.
    x = [deepcopy(x0)]
    u ->
    begin
        x_, y = stf(x[1], u)
        x[1] = x_
        return exposeState ? (y, deepcopy(x_)) : y
    end
end


"""
Evaluates state evolution of an automata defined by means of a state transition
function `stf`, starting from the given initial state `x0`, and driven by the
input sequence `inputs`.

Return an obect which can be iterated over in order to obtain sequence of states
trough which the given automaton evolves driven by the given input sequence.
"""
function evolution(automaton, inputs)
    (automaton(u) for u in inputs)
end


"""
Create a new state transition function by statically transforming the input and
the output of the given existing state transition function `stf`. The input and
output maps are `mapIn` and `mapOut`, respectively.
"""
function mapInOut(stf, mapIn, mapOut)
    (x, u) ->
    begin
        x, y = stf(x, mapIn(u))
        (x, mapOut(y))
    end
end


"""
Create a new state transition function by serially connecting two existing ones,
so that the overall input is fed to the first one, whose output is fed to the
input of the second one, which is also considered the overall output.

The joint state is a tuple of states of individual
"""
function serial(stf1, stf2)
    ((x1, x2), u) ->
    begin
        x1_, y1 = stf1(x1, u)
        x2_, y2 = stf2(x2, y1)
        ((x1_, x2_), y2)
    end
end

end
